import numpy as np
from mayavi import mlab


def sectional2nodal(x):
    return np.r_[x[0], np.convolve(x, [0.5, 0.5], "valid"), x[-1]]


def nodal2sectional(x):
    return 0.5 * (x[:-1] + x[1:])


def set_axes_equal(ax):
    """Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    """

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])


class Visualize(object):
    def __init__(self, prob):
        prob.run_model()
        self.prob = prob
        self.fig = None

    def draw_spar(self, fname="spar.png"):
        self.init_figure()

        self.draw_ocean()

        self.draw_mooring(self.prob["mooring_plot_matrix"])

        zcut = 1.0 + self.prob["main_freeboard"]
        self.draw_pontoons(self.prob["plot_matrix"], 0.5 * self.prob["fairlead_support_outer_diameter"], zcut)

        self.draw_column(
            [0.0, 0.0],
            self.prob["main_freeboard"],
            self.prob["main.section_height"],
            0.5 * self.prob["main.outer_diameter"],
            self.prob["main.stiffener_spacing"],
        )

        t_full = sectional2nodal(self.prob["main.wall_thickness"])
        self.draw_ballast(
            [0.0, 0.0],
            self.prob["main_freeboard"],
            self.prob["main.section_height"],
            0.5 * self.prob["main.outer_diameter"] - t_full,
            self.prob["main.permanent_ballast_height"],
            self.prob["variable_ballast_height"],
        )

        self.draw_column(
            [0.0, 0.0],
            self.prob["hub_height"],
            self.prob["tow.tower_section_height"],
            0.5 * self.prob["tow.tower_outer_diameter"],
            None,
            (0.9,) * 3,
        )

        if self.prob["main.buoyancy_tank_mass"] > 0.0:
            self.draw_buoyancy_tank(
                [0.0, 0.0],
                self.prob["main_freeboard"],
                self.prob["main.section_height"],
                self.prob["main.buoyancy_tank_location"],
                0.5 * self.prob["main.buoyancy_tank_diameter"],
                self.prob["main.buoyancy_tank_height"],
            )

        self.set_figure(fname)

    def draw_semi(self, fname="semi.png"):
        self.init_figure()

        self.draw_ocean()

        self.draw_mooring(self.prob["mooring_plot_matrix"])

        pontoonMat = self.prob["plot_matrix"]
        zcut = 1.0 + np.maximum(self.prob["main_freeboard"], self.prob["offset_freeboard"])
        self.draw_pontoons(pontoonMat, 0.5 * self.prob["pontoon_outer_diameter"], zcut)

        self.draw_column(
            [0.0, 0.0],
            self.prob["main_freeboard"],
            self.prob["main.section_height"],
            0.5 * self.prob["main.outer_diameter"],
            self.prob["main.stiffener_spacing"],
        )

        t_full = sectional2nodal(self.prob["main.wall_thickness"])
        self.draw_ballast(
            [0.0, 0.0],
            self.prob["main_freeboard"],
            self.prob["main.section_height"],
            0.5 * self.prob["main.outer_diameter"] - t_full,
            self.prob["main.permanent_ballast_height"],
            self.prob["variable_ballast_height"],
        )

        if self.prob["main.buoyancy_tank_mass"] > 0.0:
            self.draw_buoyancy_tank(
                [0.0, 0.0],
                self.prob["main_freeboard"],
                self.prob["main.section_height"],
                self.prob["main.buoyancy_tank_location"],
                0.5 * self.prob["main.buoyancy_tank_diameter"],
                self.prob["main.buoyancy_tank_height"],
            )

        R_semi = self.prob["radius_to_offset_column"]
        ncolumn = int(self.prob["number_of_offset_columns"])
        angles = np.linspace(0, 2 * np.pi, ncolumn + 1)
        x = R_semi * np.cos(angles)
        y = R_semi * np.sin(angles)
        for k in range(ncolumn):
            self.draw_column(
                [x[k], y[k]],
                self.prob["offset_freeboard"],
                self.prob["off.section_height"],
                0.5 * self.prob["off.outer_diameter"],
                self.prob["off.stiffener_spacing"],
            )

            t_full = sectional2nodal(self.prob["off.wall_thickness"])
            self.draw_ballast(
                [x[k], y[k]],
                self.prob["offset_freeboard"],
                self.prob["off.section_height"],
                0.5 * self.prob["off.outer_diameter"] - t_full,
                self.prob["off.permanent_ballast_height"],
                0.0,
            )

            if self.prob["off.buoyancy_tank_mass"] > 0.0:
                self.draw_buoyancy_tank(
                    [x[k], y[k]],
                    self.prob["offset_freeboard"],
                    self.prob["off.section_height"],
                    self.prob["off.buoyancy_tank_location"],
                    0.5 * self.prob["off.buoyancy_tank_diameter"],
                    self.prob["off.buoyancy_tank_height"],
                )

        self.draw_column(
            [0.0, 0.0],
            self.prob["hub_height"],
            self.prob["tow.tower_section_height"],
            0.5 * self.prob["tow.tower_outer_diameter"],
            None,
            (0.9,) * 3,
        )

        self.set_figure(fname)

    def init_figure(self):
        mysky = np.array([135, 206, 250]) / 255.0
        mysky = tuple(mysky.tolist())
        # fig = plt.figure()
        # ax = fig.add_subplot(111, projection='3d')
        # fig = mlab.figure(bgcolor=(1,)*3, size=(1600,1100))
        # fig = mlab.figure(bgcolor=mysky, size=(1600,1100))
        self.fig = mlab.figure(bgcolor=(0,) * 3, size=(1600, 1100))

    def draw_ocean(self):
        if self.fig is None:
            self.init_figure()
        npts = 100

        # mybrown = np.array([244, 170, 66]) / 255.0
        # mybrown = tuple(mybrown.tolist())
        mywater = np.array([95, 158, 160]) / 255.0  # (0.0, 0.0, 0.8) [143, 188, 143]
        mywater = tuple(mywater.tolist())
        alpha = 0.3

        # Waterplane box
        x = y = 100 * np.linspace(-1, 1, npts)
        X, Y = np.meshgrid(x, y)
        Z = np.sin(100 * X * Y)  # np.zeros(X.shape)
        # ax.plot_surface(X, Y, Z, alpha=alpha, color=mywater)
        mlab.mesh(X, Y, Z, opacity=alpha, color=mywater, figure=self.fig)

        # Sea floor
        Z = -self.prob["water_depth"] * np.ones(X.shape)
        # ax.plot_surface(10*X, 10*Y, Z, alpha=1.0, color=mybrown)
        # mlab.mesh(10*X,10*Y,Z, opacity=1.0, color=mybrown, figure=self.fig)

        # Sides
        # x = 500 * np.linspace(-1, 1, npts)
        # z = self.prob['water_depth'] * np.linspace(-1, 0, npts)
        # X,Z = np.meshgrid(x,z)
        # Y = x.max()*np.ones(Z.shape)
        ##ax.plot_surface(X, Y, Z, alpha=alpha, color=mywater)
        # mlab.mesh(X,Y,Z, opacity=alpha, color=mywater, figure=self.fig)
        # mlab.mesh(X,-Y,Z, opacity=alpha, color=mywater, figure=self.fig)
        # mlab.mesh(Y,X,Z, opacity=alpha, color=mywater, figure=self.fig)
        ##mlab.mesh(-Y,X,Z, opacity=alpha, color=mywater, figure=self.fig)

    def draw_mooring(self, mooring):
        mybrown = np.array([244, 170, 66]) / 255.0
        mybrown = tuple(mybrown.tolist())
        npts = 100

        # Sea floor
        print(self.prob["anchor_radius"])
        r = np.linspace(0, self.prob["anchor_radius"], npts)
        th = np.linspace(0, 2 * np.pi, npts)
        R, TH = np.meshgrid(r, th)
        X = R * np.cos(TH)
        Y = R * np.sin(TH)
        Z = -self.prob["water_depth"] * np.ones(X.shape)
        # ax.plot_surface(X, Y, Z, alpha=1.0, color=mybrown)
        mlab.mesh(X, Y, Z, opacity=1.0, color=mybrown, figure=self.fig)
        cmoor = (0, 0.8, 0)
        nlines = int(self.prob["number_of_mooring_connections"] * self.prob["mooring_lines_per_connection"])
        for k in range(nlines):
            # ax.plot(mooring[k,:,0], mooring[k,:,1], mooring[k,:,2], 'k', lw=2)
            mlab.plot3d(
                mooring[k, :, 0],
                mooring[k, :, 1],
                mooring[k, :, 2],
                color=cmoor,
                tube_radius=0.5 * self.prob["mooring_diameter"],
                figure=self.fig,
            )

    def draw_pontoons(self, truss, R, freeboard):
        nE = truss.shape[0]
        c = (0.5, 0, 0)
        for k in range(nE):
            if np.any(truss[k, 2, :] > freeboard):
                continue
            mlab.plot3d(truss[k, 0, :], truss[k, 1, :], truss[k, 2, :], color=c, tube_radius=R, figure=self.fig)

    def draw_column(self, centerline, freeboard, h_section, r_nodes, spacingVec=None, ckIn=None):
        npts = 20

        nsection = h_section.size
        z_nodes = np.flipud(freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))])

        th = np.linspace(0, 2 * np.pi, npts)
        for k in range(nsection):
            rk = np.linspace(r_nodes[k], r_nodes[k + 1], npts)
            z = np.linspace(z_nodes[k], z_nodes[k + 1], npts)
            R, TH = np.meshgrid(rk, th)
            Z, _ = np.meshgrid(z, th)
            X = R * np.cos(TH) + centerline[0]
            Y = R * np.sin(TH) + centerline[1]

            # Draw parameters
            if ckIn is None:
                ck = (0.6,) * 3 if np.mod(k, 2) == 0 else (0.4,) * 3
            else:
                ck = ckIn
            # ax.plot_surface(X, Y, Z, alpha=0.5, color=ck)
            mlab.mesh(X, Y, Z, opacity=0.7, color=ck, figure=self.fig)

            if spacingVec is None:
                continue

            z = z_nodes[k] + spacingVec[k]
            while z < z_nodes[k + 1]:
                rk = np.interp(z, z_nodes[k:], r_nodes[k:])
                # ax.plot(rk*np.cos(th), rk*np.sin(th), z*np.ones(th.shape), 'r', lw=0.25)
                mlab.plot3d(
                    rk * np.cos(th) + centerline[0],
                    rk * np.sin(th) + centerline[1],
                    z * np.ones(th.shape),
                    color=(0.5, 0, 0),
                    figure=self.fig,
                )
                z += spacingVec[k]

                """
                # Web
                r   = np.linspace(rk - self.prob['stiffener_web_height'][k], rk, npts)
                R, TH = np.meshgrid(r, th)
                Z, _  = np.meshgrid(z, th)
                X = R*np.cos(TH)
                Y = R*np.sin(TH)
                ax.plot_surface(X, Y, Z, alpha=0.7, color='r')

                # Flange
                r = r[0]
                h = np.linspace(0, self.prob['stiffener_flange_width'][k], npts)
                zflange = z + h - 0.5*self.prob['stiffener_flange_width'][k]
                R, TH = np.meshgrid(r, th)
                Z, _  = np.meshgrid(zflange, th)
                X = R*np.cos(TH)
                Y = R*np.sin(TH)
                ax.plot_surface(X, Y, Z, alpha=0.7, color='r')
                """

    def draw_ballast(self, centerline, freeboard, h_section, r_nodes, h_perm, h_water):
        npts = 40
        th = np.linspace(0, 2 * np.pi, npts)
        z_nodes = np.flipud(freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))])

        # Permanent ballast
        z_perm = z_nodes[0] + np.linspace(0, h_perm, npts)
        r_perm = np.interp(z_perm, z_nodes, r_nodes)
        R, TH = np.meshgrid(r_perm, th)
        Z, _ = np.meshgrid(z_perm, th)
        X = R * np.cos(TH) + centerline[0]
        Y = R * np.sin(TH) + centerline[1]
        ck = np.array([122, 85, 33]) / 255.0
        ck = tuple(ck.tolist())
        mlab.mesh(X, Y, Z, color=ck, figure=self.fig)

        # Water ballast
        z_water = z_perm[-1] + np.linspace(0, h_water, npts)
        r_water = np.interp(z_water, z_nodes, r_nodes)
        R, TH = np.meshgrid(r_water, th)
        Z, _ = np.meshgrid(z_water, th)
        X = R * np.cos(TH) + centerline[0]
        Y = R * np.sin(TH) + centerline[1]
        ck = (0.0, 0.1, 0.8)  # Dark blue
        mlab.mesh(X, Y, Z, color=ck, figure=self.fig)

    def draw_buoyancy_tank(self, centerline, freeboard, h_section, loc, r_box, h_box):
        npts = 20
        z_nodes = np.flipud(freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))])
        z_lower = loc * (z_nodes[-1] - z_nodes[0]) + z_nodes[0]

        # Lower and Upper surfaces
        r = np.linspace(0, r_box, npts)
        th = np.linspace(0, 2 * np.pi, npts)
        R, TH = np.meshgrid(r, th)
        X = R * np.cos(TH) + centerline[0]
        Y = R * np.sin(TH) + centerline[1]
        Z = z_lower * np.ones(X.shape)
        ck = (0.9,) * 3
        mlab.mesh(X, Y, Z, opacity=0.5, color=ck, figure=self.fig)
        Z += h_box
        mlab.mesh(X, Y, Z, opacity=0.5, color=ck, figure=self.fig)

        # Cylinder part
        z = z_lower + np.linspace(0, h_box, npts)
        Z, TH = np.meshgrid(z, th)
        R = r_box * np.ones(Z.shape)
        X = R * np.cos(TH) + centerline[0]
        Y = R * np.sin(TH) + centerline[1]
        mlab.mesh(X, Y, Z, opacity=0.5, color=ck, figure=self.fig)

    def set_figure(self, fname=None):
        # ax.set_aspect('equal')
        # set_axes_equal(ax)
        # ax.autoscale_view(tight=True)
        # ax.set_xlim([-125, 125])
        # ax.set_ylim([-125, 125])
        # ax.set_zlim([-220, 30])
        # plt.axis('off')
        # plt.show()
        # mlab.move([-517.16728532, -87.0711504, 5.60826224], [1.35691603e+01, -2.84217094e-14, -1.06547500e+02])
        # mlab.view(-170.68320804213343, 78.220729198686854, 549.40101471336777, [1.35691603e+01,  0.0, -1.06547500e+02])
        if not fname is None:
            fpart = fname.split(".")
            if len(fpart) == 1 or not fpart[-1].lower() in ["jpg", "png", "bmp"]:
                fname += ".png"
            mlab.savefig(fname, figure=self.fig)
        mlab.show()



"""

"""

from marmot import Object


class Cargo(Object):
    def __repr__(self):
        return self.type

    @property
    def type(self):
        """Returns type of `Cargo`."""
        return self.__class__.__name__



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2021, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import false, process

from wisdem.orbit.core import Cargo
from wisdem.orbit.core.defaults import process_times as pt


class Jacket(Cargo):
    """Jacket Cargo"""

    def __init__(
        self,
        height=None,
        mass=None,
        deck_space=None,
        foundation_type="piles",
        **kwargs,
    ):
        """Creats an instance of `Jacket`."""

        self.height = height
        self.mass = mass
        self.deck_space = deck_space
        self.num_legs = kwargs.get("num_legs", 4)
        self.foundation_type = foundation_type

    @staticmethod
    def fasten(**kwargs):
        """Returns time requred to fasten a jacket at port."""

        key = "jacket_fasten_time"
        time = kwargs.get(key, pt[key])

        return "Fasten Jacket", time

    @staticmethod
    def release(**kwargs):
        """Returns time required to release jacket from fastenings."""

        key = "jacket_release_time"
        time = kwargs.get(key, pt[key])

        return "Release Jacket", time


@process
def install_piles(vessel, jacket, **kwargs):
    """
    Process logic for installing piles at site.

    Parameters
    ----------
    vessel : Vessel
    jacket : dict
    """

    reposition_time = kwargs.get("jacket_vessel_reposition", pt["jacket_vessel_reposition"])
    position_pile_time = kwargs.get("jacket_position_pile", pt["jacket_position_pile"])
    drive_time = kwargs.get("jacket_pile_drive_time", pt["jacket_pile_drive_time"])
    pin_template_time = kwargs.get("jacket_pin_template_time", pt["jacket_pin_template_time"])

    yield vessel.task_wrapper(
        "Lay Pin Template",
        pin_template_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )

    for i in range(jacket.num_legs):
        yield vessel.task_wrapper(
            "Position Pile",
            position_pile_time,
            constraints={**vessel.operational_limits},
            **kwargs,
        )

        yield vessel.task_wrapper(
            "Drive Pile",
            drive_time,
            constraints={**vessel.operational_limits, "night": false()},
            suspendable=True,
            **kwargs,
        )

        if i < (jacket.num_legs - 1):
            yield vessel.task_wrapper(
                "Move to Next Leg",
                reposition_time,
                constraints=vessel.transit_limits,
                suspendable=True,
                **kwargs,
            )


@process
def install_suction_buckets(vessel, jacket, **kwargs):
    """
    Process logic for installing suction buckets at site.

    Parameters
    ----------
    vessel : Vessel
    jacket : dict
    """

    reposition_time = kwargs.get("jacket_vessel_reposition", pt["jacket_vessel_reposition"])
    install_time = kwargs.get("jacket_suction_bucket", pt["jacket_suction_bucket"])

    for i in range(jacket.num_legs):
        yield vessel.task_wrapper(
            "Install Suction Bucket",
            install_time,
            constraints={**vessel.operational_limits},
            **kwargs,
        )

        if i < (jacket.num_legs - 1):
            yield vessel.task_wrapper(
                "Move to Next Leg",
                reposition_time,
                constraints=vessel.transit_limits,
                suspendable=True,
                **kwargs,
            )


@process
def install_jacket(vessel, jacket, **kwargs):
    """
    Process logic for installing a jacket at site.

    Parameters
    ----------
    vessel : Vessel
    jacket : dict
    """

    if jacket.foundation_type == "piles":
        yield install_piles(vessel, jacket, **kwargs)

    elif jacket.foundation_type == "suction":
        yield install_suction_buckets(vessel, jacket, **kwargs)

    else:
        return ValueError("Input 'jacket.foundation_type' must be 'piles' or 'suction'.")

    lift_time = kwargs.get("jacket_lift_time", pt["jacket_lift_time"])
    yield vessel.task_wrapper(
        "Lift Jacket",
        lift_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )

    lower_time = kwargs.get("jacket_lower_time", pt["jacket_lower_time"])
    yield vessel.task_wrapper(
        "Lower and Position Jacket",
        lower_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )

    grout_time = kwargs.get("jacket_grout_time", pt["jacket_grout_time"])
    yield vessel.task_wrapper("Grout Jacket", grout_time, constraints=vessel.transit_limits, **kwargs)



from typing import Dict, Any, List, Union
import yaml
import sys, traceback
import re
from pathlib import Path

from PySide2.QtWidgets import (  # type: ignore
    QHBoxLayout,
    QVBoxLayout,
    QLineEdit,
    QPushButton,
    QLabel,
    QFormLayout,
    QTabWidget,
    QWidget,
    QMainWindow,
    QFileDialog,
    QApplication,
    QMessageBox,
)

from wisdem.glue_code.runWISDEM import run_wisdem


class FocusQLineEdit(QLineEdit):
    """
    FocusQLineEdit subclasses QLineEdit to add the following functionality:

    - Observing focus loss events and, when focus is lost

    - Updating a value in given dictionary and key based on the input
      in this text field.
    """

    def __init__(self, *args, **kwargs):
        """
        The instance attributes for this class are:

        _dictionary: Dict[str, Any]
            The dictionary to be changed when this text field is changed.

        _key_on_dictionary: str
            The key on the dictionary to be changed when this text field
            changes.

        _list_re: re.Pattern
            The regex that matches a string that contains a list, so that
            it does not need to be instantiated each time it is used.

        These attributes default to None. If either one of them is None, no
        attempt will be made to change the dictionary.
        """
        self._dictionary = None
        self._key_on_dictionary = None
        self._list_re = re.compile(r"^\[.*\]$")
        super(FocusQLineEdit, self).__init__(*args, *kwargs)

    def set_dictionary_and_key(self, dictionary: Union[List[Any], Dict[str, Any]], key: str) -> None:
        """
        This method sets the dictionary and key to be modified when the focus
        changes out of this widget

        Parameters
        ----------
        dictionary: Dict[str, Any]
            The dictionary to be changed.

        key: str
            The key whose value is to be changed on the dictionary.
        """
        self._dictionary = dictionary
        self._key_on_dictionary = key

    def focusOutEvent(self, arg__1) -> None:
        """
        Overrides focusOutEvent() in the base class (the base class's method is
        called before anything in this method executes, though).

        The purpose of this override is to observe focus out events. When this
        control looses focus, it updates the underlying specified in the
        instance attributes.
        """
        super(FocusQLineEdit, self).focusOutEvent(arg__1)
        if self._dictionary is not None and self._key_on_dictionary is not None:
            if self.is_list(self.text()):
                value = self.parse_list()
            elif self.is_float(self.text()):
                value = float(self.text())  # type: ignore
            elif self.is_boolean(self.text()):
                value = self.parse_boolean(self.text())  # type: ignore
            else:
                value = self.text()  # type: ignore
            self._dictionary[self._key_on_dictionary] = value
            # print("New level dictionary", self._dictionary)
        # else:
        #     print("Focus lost, but dictionary and key are not set.")

    def parse_list(self) -> List[Union[str, float]]:
        """
        This parses the text in the field to a list of numbers and/or strings.

        Returns
        -------
        List[Union[str, float]]
            A list that contains strings and floats, depending on what could be
            parsed out of the text in the field.
        """
        trimmed_text = self.text()[1:-1]
        trimmed_str_values = [x.strip() for x in trimmed_text.split(",")]
        result: List[Union[str, float]] = []
        for x in trimmed_str_values:
            if self.is_integer(x):
                result.append(int(x))
            elif self.is_float(x):
                result.append(float(x))
            else:
                result.append(x)
        return result

    @staticmethod
    def parse_boolean(value: str) -> bool:
        """
        This method parses a string as a boolean. See the is_boolean()
        method below.

        Parameters
        ----------
        value: str
            The value that is being parsed as a boolean.

        Returns
        -------
        bool
            The value, parsed as a boolean.
        """
        if value == "True" or value == "true":
            return True
        else:
            return False

    @staticmethod
    def is_float(value: Any) -> bool:
        """
        This tests if a value is a float and returns True or False depending
        on the outcome of the test

        Parameters
        ----------
        value: Any
            The value to test.

        Returns
        -------
        bool
            True if the value is a float, False otherwise.
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_integer(value: Any) -> bool:
        """
        This tests if a value is an integer and returns True or False depending
        on the outcome of the test

        Parameters
        ----------
        value: Any
            The value to test.

        Returns
        -------
        bool
            True if the value is an integer, False otherwise.
        """
        try:
            int(value)
            return True
        except ValueError:
            return False

    def is_list(self, value: str) -> bool:
        """
        Determines whether a value is a list through a regular expression
        match. The intent is that, if the value is a list, it will be
        parsed as a list.

        Parameters
        ----------
        value: str

        Returns
        -------
        bool
            True if the value appears to be a list, false otherwise
        """
        return self._list_re.match(value) is not None

    def is_boolean(self, value: str) -> bool:
        """
        Determines whether a string encodes a "True" or a "False" value. It
        does not parse the boolean value. If this method returns True, then
        the value can be parsed as a True or False boolean.

        Both Pythonic True and False and JSON-style true and false are
        accepted as booleans.

        Note: This should probably be checkbox, not a text box, but this
        is a temporary workaround.

        Parameters
        ----------
        value: str
            The string to be tested.

        Returns
        -------
        bool
            True if the string represents a boolean, false otherwise.
        """
        return value == "True" or value == "False" or value == "true" or value == "false"


class FormAndMenuWindow(QMainWindow):
    """
    This class creates a form to edit a dictionary. It nests tabs for different
    levels of the nesting within the dictionaries.

    This automatically builds an interface from the dictionary.
    """

    def __init__(self, parent=None):
        """
        Parameters
        ----------
        dict_to_edit: Dict[str, Any]
            The dictionary to edit in this form.

        output_filename: str
            The filename to write the dictionary when the save button is clicked.

        parent
            The parent as needed by the base class.
        """
        super(FormAndMenuWindow, self).__init__(parent)
        self.analysis_yaml_editor_widget = None
        self.modeling_yaml_widget = None
        self.geometry_yaml_widget = None
        self.geometry_filename_line_edit = None
        self.modeling_filename_line_edit = None
        self.analysis_filename_line_edit = None
        self.geometry_dict = None
        self.analysis_dict = None
        self.modeling_dict = None
        self.geometry_filename = None
        self.analysis_filename = None
        self.modeling_filename = None
        self.main_widget = None
        self.status_widget = None
        self.status_label = None

    def setup(self) -> None:
        """
        After this class is instantiated, this method should be called to
        lay out the user interface.
        """
        self.setWindowTitle("YAML GUI")
        # self.setup_menu_bar()

        central_widget = self.create_central_widget()
        self.setCentralWidget(central_widget)

    def recursion_ui_setup(self, dict_or_list: Union[List[Any], Dict[str, Any]]) -> QFormLayout:
        """
        This recursive method is where the automatic layout magic happens.
        This method calls itself recursively as it descends down the dictionary
        nesting structure.

        Basically, any given dictionary can consist of scalar and dictionary
        values. At each level of the dictionary, edit fields are placed for
        scalar values and tabbed widgets are placed for the next level of
        nesting.

        Parameters
        ----------
        dict_or_list: Dict[str, Any]
            The dictionary to automatically lay out in to the interface.
        """
        form_level_layout = QFormLayout()
        dict_tabs = QTabWidget()
        display_tabs = False
        subscripts_values = dict_or_list.items() if type(dict_or_list) is dict else enumerate(dict_or_list)  # type: ignore
        for k, v in subscripts_values:

            # Recursive call for nested dictionaries within dictionaries.
            if type(v) is dict:
                display_tabs = True
                child_widget = QWidget()
                child_layout = self.recursion_ui_setup(v)
                child_widget.setLayout(child_layout)
                dict_tabs.addTab(child_widget, str(k))

            # Recursive call for nested dictionaries within lists.
            elif type(v) is list and type(v[0]) is dict:
                display_tabs = True
                child_widget = QWidget()
                child_layout = self.recursion_ui_setup(v)
                child_widget.setLayout(child_layout)
                dict_tabs.addTab(child_widget, str(k))

            # Otherwise just lay out a label and text field.
            else:
                line_edit = FocusQLineEdit(str(v))
                line_edit.setMinimumWidth(150)
                line_edit.set_dictionary_and_key(dict_or_list, k)
                form_level_layout.addRow(QLabel(k), line_edit)

        # If there is a nested dictionary, display it.
        if display_tabs:
            form_level_layout.addRow(dict_tabs)

        # Return the whole layout
        return form_level_layout

    def write_dict_to_yaml(self) -> None:
        """
        This is the event handler for the save button. It simply writes the
        dictionary (which has been continuously updated during focus out
        events) to a YAML file as specified in self.output_filename.
        """
        with open(self.output_filename, "w") as file:
            yaml.dump(self.dict_to_edit, file)

    def create_central_widget(self) -> QWidget:
        """
        Returns
        -------
        QWidget
            The form with buttons on it.
        """
        status_and_main_widget = QWidget()
        status_and_main_widget_layout = QVBoxLayout()

        self.status_widget = QWidget()
        status_form_layout = QFormLayout()
        status_label = QLabel("Status:")
        self.status_label = QLabel()
        self.status_label.setText("Please create simulation configurations.")
        status_form_layout.addRow(status_label, self.status_label)
        self.status_widget.setLayout(status_form_layout)

        self.main_widget = QWidget()
        subsection_width = 500
        subsection_height = 900

        geometry_section_label = QLabel("Geometry")
        geometry_section_label.setStyleSheet("font-weight: bold;")
        geometry_filename_button = QPushButton("Select geometry YAML...")
        # geometry_visualize_button = QPushButton("Visualize geometry")
        self.geometry_filename_line_edit = QLineEdit()
        self.geometry_filename_line_edit.setPlaceholderText("Please select a geometry file.")
        self.geometry_filename_line_edit.setFixedWidth(200)
        self.geometry_filename_line_edit.setReadOnly(True)
        geometry_filename_button.clicked.connect(self.file_picker_geometry)

        modeling_section_label = QLabel("Modeling")
        modeling_section_label.setStyleSheet("font-weight: bold;")
        modeling_filename_button = QPushButton("Select modeling YAML...")
        self.modeling_filename_line_edit = QLineEdit()
        self.modeling_filename_line_edit.setPlaceholderText("Please select a modeling file.")
        self.modeling_filename_line_edit.setFixedWidth(200)
        self.modeling_filename_line_edit.setReadOnly(True)
        modeling_filename_button.clicked.connect(self.file_picker_modeling)

        analysis_section_label = QLabel("Analysis")
        analysis_section_label.setStyleSheet("font-weight: bold;")
        analysis_filename_button = QPushButton("Select analysis YAML...")
        self.analysis_filename_line_edit = QLineEdit()
        self.analysis_filename_line_edit.setPlaceholderText("Please select an analysis file...")
        self.analysis_filename_line_edit.setFixedWidth(200)
        self.analysis_filename_line_edit.setReadOnly(True)
        analysis_filename_button.clicked.connect(self.file_picker_analysis)

        run_weis_button = QPushButton("Run WISDEM")
        run_weis_button.clicked.connect(self.run_weis_clicked)

        self.modeling_yaml_widget = QWidget()
        self.analysis_yaml_editor_widget = QWidget()
        self.geometry_yaml_widget = QWidget()

        geometry_layout = QFormLayout()
        geometry_layout.addRow(geometry_section_label)
        geometry_layout.addRow(self.geometry_filename_line_edit, geometry_filename_button)
        # geometry_layout.addRow(geometry_visualize_button)
        geometry_layout.addRow(self.geometry_yaml_widget)
        geometry_widget = QWidget()
        geometry_widget.setFixedWidth(subsection_width)
        geometry_widget.setFixedHeight(subsection_height)
        geometry_widget.setLayout(geometry_layout)

        modeling_layout = QFormLayout()
        modeling_layout.addRow(modeling_section_label)
        modeling_layout.addRow(self.modeling_filename_line_edit, modeling_filename_button)
        modeling_layout.addRow(self.modeling_yaml_widget)
        modeling_widget = QWidget()
        modeling_widget.setFixedWidth(subsection_width)
        modeling_widget.setFixedHeight(subsection_height)
        modeling_widget.setLayout(modeling_layout)

        analysis_layout = QFormLayout()
        analysis_layout.addRow(analysis_section_label)
        analysis_layout.addRow(self.analysis_filename_line_edit, analysis_filename_button)
        analysis_layout.addRow(self.analysis_yaml_editor_widget)
        analysis_widget = QWidget()
        analysis_widget.setFixedWidth(subsection_width)
        analysis_widget.setFixedHeight(subsection_height)
        analysis_widget.setLayout(analysis_layout)

        main_layout = QHBoxLayout()
        main_layout.addWidget(geometry_widget)
        main_layout.addWidget(modeling_widget)
        main_layout.addWidget(analysis_widget)
        main_layout.addWidget(run_weis_button)

        self.main_widget.setLayout(main_layout)
        # return self.main_widget

        status_and_main_widget_layout.addWidget(self.status_widget)
        status_and_main_widget_layout.addWidget(self.main_widget)
        status_and_main_widget.setLayout(status_and_main_widget_layout)

        return status_and_main_widget

    def run_weis_clicked(self):
        """
        When the "Run WEIS" button is clicked, a popup dialog
        will pop up asking what the user wants to do. If the user
        has skipped any configuration files, then it will tell them
        to go back and create those configuration files. If the user
        has specified all the configuration files, then it will prompt
        the user for confirmation before saving the WEIS configuration
        files.
        """
        if self.geometry_filename is None:
            msg = QMessageBox()
            msg.setText("Run WISDEM: Missing file")
            msg.setInformativeText("You did not specify a geometry file.")
            msg.addButton(QMessageBox.Cancel)
            msg.exec()
        elif self.modeling_filename is None:
            msg = QMessageBox()
            msg.setText("Run WISDEM: Missing file")
            msg.setInformativeText("You did not specify a modeling file.")
            msg.addButton(QMessageBox.Cancel)
            msg.exec()
        elif self.analysis_filename is None:
            msg = QMessageBox()
            msg.setText("Run WISDEM: Missing file")
            msg.setInformativeText("You did not specify an analysis file.")
            msg.addButton(QMessageBox.Cancel)
            msg.exec()
        else:
            self.status_label.setText("Writing files...")
            self.write_configuration_files()
            self.status_label.setText("Configuration files written.")
            msg = QMessageBox()
            msg.setText("Run WISDEM: Configuration files complete!")
            msg.setInformativeText("Click cancel to back out and continue editing. Click OK to run WISDEM.")
            msg.addButton(QMessageBox.Cancel)
            msg.addButton(QMessageBox.Ok)
            choice = msg.exec()
            if choice == QMessageBox.Ok:
                self.disable_ui_and_execute_wisdem()

    def write_configuration_files(self):
        """
        The "Run WEIS" click event handler calls this method when it is ready
        to make an attempt to run WEIS. It checks to see if all file shave been
        edited
        """
        if self.geometry_filename is not None:
            print(f"Writing geometry: {self.geometry_filename}")
            with open(self.geometry_filename, "w") as file:
                yaml.dump(self.geometry_dict, file)
        else:
            print("No geometry file to write")

        if self.analysis_filename is not None:
            print(f"Writing analysis: {self.analysis_filename}")
            with open(self.analysis_filename, "w") as file:
                yaml.dump(self.analysis_dict, file)
        else:
            print("No analysis file to write")

        if self.modeling_filename is not None:
            print(f"Writing modeling: {self.modeling_filename}")
            with open(self.modeling_filename, "w") as file:
                yaml.dump(self.modeling_dict, file)
        else:
            print("No modeling file to write")

    def disable_ui_and_execute_wisdem(self):
        """
        This method disables all widgets on the UI and executes WISDEM. It
        displays message boxes depending on whether WISDEM executed
        successfully.
        """
        self.main_widget.setEnabled(False)
        self.status_label.setText("Running WISDEM")

        try:
            wt_opt, modeling_options, analysis_options = run_wisdem(
                self.geometry_filename, self.modeling_filename, self.analysis_filename
            )
        except Exception as err:
            short_error_message = f"{type(err)}: {err}. More details on command line."
            traceback.print_exc(file=sys.stdout)
            self.status_label.setText("Execution error")
            msg = QMessageBox()
            msg.setText("WISDEM execution error")
            msg.setInformativeText(short_error_message)
            msg.addButton(QMessageBox.Ok)
            msg.exec()
            self.main_widget.setEnabled(True)
        else:
            self.status_label.setText("Execution success")
            msg = QMessageBox()
            msg.setText("WISDEM executed successfully")
            msg.addButton(QMessageBox.Ok)
            msg.exec()
            self.main_widget.setEnabled(True)

    def file_picker_geometry(self):
        """
        Shows the open file dialog for the geometry file.

        Returns
        -------
        None
            Returns nothing for now.
        """
        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setViewMode(QFileDialog.Detail)
        self.geometry_filename, _ = dialog.getOpenFileName(None, "Open File", str(Path.home()), "YAML (*.yml *.yaml)")
        self.geometry_filename_line_edit.setText(self.geometry_filename)
        self.geometry_dict = self.read_yaml_to_dictionary(self.geometry_filename)
        layout = self.recursion_ui_setup(self.geometry_dict)
        self.geometry_yaml_widget.setLayout(layout)

    def file_picker_modeling(self):
        """
        Shows the open dialog

        Returns
        -------
        None
            Returns nothing for now.
        """
        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setViewMode(QFileDialog.Detail)
        self.modeling_filename, _ = dialog.getOpenFileName(None, "Open File", str(Path.home()), "YAML (*.yml *.yaml)")
        self.modeling_filename_line_edit.setText(self.modeling_filename)
        self.modeling_dict = self.read_yaml_to_dictionary(self.modeling_filename)
        layout = self.recursion_ui_setup(self.modeling_dict)
        self.modeling_yaml_widget.setLayout(layout)

    def file_picker_analysis(self) -> None:
        """
        Shows the open dialog for the analysis YAML

        Returns
        -------
        None
        """
        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setViewMode(QFileDialog.Detail)
        self.analysis_filename, _ = dialog.getOpenFileName(None, "Open File", str(Path.home()), "YAML (*.yml *.yaml)")
        self.analysis_filename_line_edit.setText(self.analysis_filename)
        self.analysis_dict = self.read_yaml_to_dictionary(self.analysis_filename)
        layout = self.recursion_ui_setup(self.analysis_dict)
        self.analysis_yaml_editor_widget.setLayout(layout)

    @staticmethod
    def read_yaml_to_dictionary(input_filename: str) -> Dict[str, Any]:
        """
        This reads the YAML input which is used to build the user interface.
        """
        with open(input_filename) as file:
            result = yaml.load(file, Loader=yaml.FullLoader)
        return result


def run():
    # Create the Qt Application
    app = QApplication(sys.argv)
    # Create and show the form
    form = FormAndMenuWindow()
    form.setup()
    form.show()
    # Run the main Qt loop
    sys.exit(app.exec_())



import numpy as np
from wisdem.commonse.constants import eps
from wisdem.commonse.utilities import smooth_max, smooth_min, cubic_spline_eval


def hoopStressEurocode(d, t, L_reinforced, hoop):
    """default method for scaling hoop stress using Eurocode method
    GB 06/21/2018: Ansys comparisons for submerged case suggests this over-compensates for stiffener
                   I'm not even sure the Eurocode is implemented correctly here.  Suggest using the standard
                   hoop stress expression above or API's handling of ring stiffeners below.
    """

    r = d / 2.0 - t / 2.0  # radius of cylinder middle surface
    omega = L_reinforced / np.sqrt(r * t)

    C_theta = 1.5  # clamped-clamped
    k_w = 0.46 * (1.0 + 0.1 * np.sqrt(C_theta / omega * r / t))
    k_w, _, _ = smooth_max(k_w, 0.65)
    k_w, _, _ = smooth_min(k_w, 1.0)
    return k_w * hoop


def bucklingGL(d, t, Fz, Myy, tower_height, E, sigma_y, gamma_f=1.2, gamma_b=1.1):

    # other factors
    alpha = 0.21  # buckling imperfection factor
    beta = 1.0  # bending coefficient
    sk_factor = 2.0  # fixed-free
    tower_height = tower_height * sk_factor

    # geometry
    A = np.pi * d * t
    I = np.pi * (d / 2.0) ** 3 * t
    Wp = I / (d / 2.0)

    # applied loads
    Nd = -Fz * gamma_f
    Md = Myy * gamma_f

    # plastic resistance
    Np = A * sigma_y / gamma_b
    Mp = Wp * sigma_y / gamma_b

    # factors
    Ne = np.pi ** 2 * (E * I) / (1.1 * tower_height ** 2)
    lambda_bar = np.sqrt(Np * gamma_b / Ne)
    phi = 0.5 * (1 + alpha * (lambda_bar - 0.2) + lambda_bar ** 2)
    kappa = np.ones(A.shape)
    idx = lambda_bar > 0.2
    kappa[idx] = 1.0 / (phi[idx] + np.sqrt(phi[idx] ** 2 - lambda_bar[idx] ** 2))
    delta_n = 0.25 * kappa * lambda_bar ** 2
    delta_n = np.minimum(delta_n, 0.1)

    GL_utilization = Nd / (kappa * Np) + beta * Md / Mp + delta_n  # this is utilization must be <1

    return GL_utilization


def shellBucklingEurocode(d, t, sigma_z, sigma_t, tau_zt, L_reinforced, E, sigma_y, gamma_f=1.2, gamma_b=1.1):
    """
    Estimate shell buckling utilization along tower.

    Arguments:
    npt - number of locations at each node at which stress is evaluated.
    sigma_z - axial stress at npt*node locations.  must be in order
                  [(node1_pts1-npt), (node2_pts1-npt), ...]
    sigma_t - azimuthal stress given at npt*node locations
    tau_zt - shear stress (z, theta) at npt*node locations
    E - modulus of elasticity
    sigma_y - yield stress
    L_reinforced - reinforcement length - structure is re-discretized with this spacing
    gamma_f - safety factor for stresses
    gamma_b - safety factor for buckling

    Returns:
    z
    EU_utilization: - array of shell buckling utilizations evaluted at (z[0] at npt locations, \n
                      z[0]+L_reinforced at npt locations, ...). \n
                      Each utilization must be < 1 to avoid failure.
    """

    n = len(t)
    EU_utilization = np.zeros(n)
    sigma_z_sh = np.zeros(n)
    sigma_t_sh = np.zeros(n)
    tau_zt_sh = np.zeros(n)

    for i in range(n):
        h = L_reinforced[i]

        r1 = d[i] / 2.0 - t[i] / 2.0
        r2 = d[i + 1] / 2.0 - t[i] / 2.0

        sigma_z_shell = sigma_z[i]
        sigma_t_shell = sigma_t[i]
        tau_zt_shell = tau_zt[i]

        # TODO: the following is non-smooth, although in general its probably OK
        # change to magnitudes and add safety factor
        sigma_z_shell = gamma_f * abs(sigma_z_shell)
        sigma_t_shell = gamma_f * abs(sigma_t_shell)
        tau_zt_shell = gamma_f * abs(tau_zt_shell)

        EU_utilization[i] = _shellBucklingOneSection(
            h, r1, r2, t[i], gamma_b, sigma_z_shell, sigma_t_shell, tau_zt_shell, E[i], sigma_y[i]
        )

        # make them into vectors
        sigma_z_sh[i] = sigma_z_shell
        sigma_t_sh[i] = sigma_t_shell
        tau_zt_sh[i] = tau_zt_shell

    return EU_utilization  # this is utilization must be <1


def _cxsmooth(omega, rovert):

    Cxb = 6.0  # clamped-clamped
    constant = 1 + 1.83 / 1.7 - 2.07 / 1.7 ** 2

    ptL1 = 1.7 - 0.25
    ptR1 = 1.7 + 0.25

    ptL2 = 0.5 * rovert - 1.0
    ptR2 = 0.5 * rovert + 1.0

    ptL3 = (0.5 + Cxb) * rovert - 1.0
    ptR3 = (0.5 + Cxb) * rovert + 1.0

    if omega < ptL1:
        Cx = constant - 1.83 / omega + 2.07 / omega ** 2

    elif omega >= ptL1 and omega <= ptR1:

        fL = constant - 1.83 / ptL1 + 2.07 / ptL1 ** 2
        fR = 1.0
        gL = 1.83 / ptL1 ** 2 - 4.14 / ptL1 ** 3
        gR = 0.0
        Cx = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        Cx = 1.0

    elif omega >= ptL2 and omega <= ptR2:

        fL = 1.0
        fR = 1 + 0.2 / Cxb * (1 - 2.0 * ptR2 / rovert)
        gL = 0.0
        gR = -0.4 / Cxb / rovert
        Cx = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    elif omega > ptR2 and omega < ptL3:
        Cx = 1 + 0.2 / Cxb * (1 - 2.0 * omega / rovert)

    elif omega >= ptL3 and omega <= ptR3:

        fL = 1 + 0.2 / Cxb * (1 - 2.0 * ptL3 / rovert)
        fR = 0.6
        gL = -0.4 / Cxb / rovert
        gR = 0.0
        Cx = cubic_spline_eval(ptL3, ptR3, fL, fR, gL, gR, omega)

    else:
        Cx = 0.6

    return Cx


def _sigmasmooth(omega, E, rovert):

    Ctheta = 1.5  # clamped-clamped

    ptL = 1.63 * rovert * Ctheta - 1
    ptR = 1.63 * rovert * Ctheta + 1

    if omega < 20.0 * Ctheta:
        offset = 10.0 / (20 * Ctheta) ** 2 - 5 / (20 * Ctheta) ** 3
        Cthetas = 1.5 + 10.0 / omega ** 2 - 5 / omega ** 3 - offset
        sigma = 0.92 * E * Cthetas / omega / rovert

    elif omega >= 20.0 * Ctheta and omega < ptL:

        sigma = 0.92 * E * Ctheta / omega / rovert

    elif omega >= ptL and omega <= ptR:

        alpha1 = 0.92 / 1.63 - 2.03 / 1.63 ** 4

        fL = 0.92 * E * Ctheta / ptL / rovert
        fR = E * (1.0 / rovert) ** 2 * (alpha1 + 2.03 * (Ctheta / ptR * rovert) ** 4)
        gL = -0.92 * E * Ctheta / rovert / ptL ** 2
        gR = -E * (1.0 / rovert) * 2.03 * 4 * (Ctheta / ptR * rovert) ** 3 * Ctheta / ptR ** 2

        sigma = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, omega)

    else:

        alpha1 = 0.92 / 1.63 - 2.03 / 1.63 ** 4
        sigma = E * (1.0 / rovert) ** 2 * (alpha1 + 2.03 * (Ctheta / omega * rovert) ** 4)

    return sigma


def _tausmooth(omega, rovert):

    ptL1 = 9
    ptR1 = 11

    ptL2 = 8.7 * rovert - 1
    ptR2 = 8.7 * rovert + 1

    if omega < ptL1:
        C_tau = np.sqrt(1.0 + 42.0 / omega ** 3 - 42.0 / 10 ** 3)

    elif omega >= ptL1 and omega <= ptR1:
        fL = np.sqrt(1.0 + 42.0 / ptL1 ** 3 - 42.0 / 10 ** 3)
        fR = 1.0
        gL = -63.0 / ptL1 ** 4 / fL
        gR = 0.0
        C_tau = cubic_spline_eval(ptL1, ptR1, fL, fR, gL, gR, omega)

    elif omega > ptR1 and omega < ptL2:
        C_tau = 1.0

    elif omega >= ptL2 and omega <= ptR2:
        fL = 1.0
        fR = 1.0 / 3.0 * np.sqrt(ptR2 / rovert) + 1 - np.sqrt(8.7) / 3
        gL = 0.0
        gR = 1.0 / 6 / np.sqrt(ptR2 * rovert)
        C_tau = cubic_spline_eval(ptL2, ptR2, fL, fR, gL, gR, omega)

    else:
        C_tau = 1.0 / 3.0 * np.sqrt(omega / rovert) + 1 - np.sqrt(8.7) / 3

    return C_tau


def _shellBucklingOneSection(h, r1, r2, t, gamma_b, sigma_z, sigma_t, tau_zt, E, sigma_y):
    """
    Estimate shell buckling for one tapered cylindrical shell section.

    Arguments:
    h - height of conical section
    r1 - radius at bottom
    r2 - radius at top
    t - shell thickness
    E - modulus of elasticity
    sigma_y - yield stress
    gamma_b - buckling reduction safety factor
    sigma_z - axial stress component
    sigma_t - azimuthal stress component
    tau_zt - shear stress component (z, theta)

    Returns:
    EU_utilization, shell buckling utilization which must be < 1 to avoid failure

    """

    # NOTE: definition of r1, r2 switched from Eurocode document to be consistent with FEM.

    # ----- geometric parameters --------
    beta = np.arctan2(r1 - r2, h)
    L = h / np.cos(beta)

    # ------------- axial stress -------------
    # length parameter
    le = L
    re = 0.5 * (r1 + r2) / np.cos(beta)
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # compute Cx
    Cx = _cxsmooth(omega, rovert)

    # if omega <= 1.7:
    #     Cx = 1.36 - 1.83/omega + 2.07/omega/omega
    # elif omega > 0.5*rovert:
    #     Cxb = 6.0  # clamped-clamped
    #     Cx = max(0.6, 1 + 0.2/Cxb*(1-2.0*omega/rovert))
    # else:
    #     Cx = 1.0

    # critical axial buckling stress
    sigma_z_Rcr = 0.605 * E * Cx / rovert

    # compute buckling reduction factors
    lambda_z0 = 0.2
    beta_z = 0.6
    eta_z = 1.0
    Q = 25.0  # quality parameter - high
    lambda_z = np.sqrt(sigma_y / sigma_z_Rcr)
    delta_wk = 1.0 / Q * np.sqrt(rovert) * t
    alpha_z = 0.62 / (1 + 1.91 * (delta_wk / t) ** 1.44)

    chi_z = _buckling_reduction_factor(alpha_z, beta_z, eta_z, lambda_z0, lambda_z)

    # design buckling stress
    sigma_z_Rk = chi_z * sigma_y
    sigma_z_Rd = sigma_z_Rk / gamma_b

    # ---------------- hoop stress ------------------

    # length parameter
    le = L
    re = 0.5 * (r1 + r2) / (np.cos(beta))
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # Ctheta = 1.5  # clamped-clamped
    # CthetaS = 1.5 + 10.0/omega**2 - 5.0/omega**3

    # # critical hoop buckling stress
    # if (omega/Ctheta < 20.0):
    #     sigma_t_Rcr = 0.92*E*CthetaS/omega/rovert
    # elif (omega/Ctheta > 1.63*rovert):
    #     sigma_t_Rcr = E*(1.0/rovert)**2*(0.275 + 2.03*(Ctheta/omega*rovert)**4)
    # else:
    #     sigma_t_Rcr = 0.92*E*Ctheta/omega/rovert

    sigma_t_Rcr = np.maximum(eps, _sigmasmooth(omega, E, rovert))

    # buckling reduction factor
    alpha_t = 0.65  # high fabrication quality
    lambda_t0 = 0.4
    beta_t = 0.6
    eta_t = 1.0
    lambda_t = np.sqrt(sigma_y / sigma_t_Rcr)

    chi_theta = _buckling_reduction_factor(alpha_t, beta_t, eta_t, lambda_t0, lambda_t)

    sigma_t_Rk = chi_theta * sigma_y
    sigma_t_Rd = sigma_t_Rk / gamma_b

    # ----------------- shear stress ----------------------

    # length parameter
    le = h
    rho = np.sqrt((r1 + r2) / (2.0 * r2))
    re = (1.0 + rho - 1.0 / rho) * r2 * np.cos(beta)
    omega = le / np.sqrt(re * t)
    rovert = re / t

    # if (omega < 10):
    #     C_tau = np.sqrt(1.0 + 42.0/omega**3)
    # elif (omega > 8.7*rovert):
    #     C_tau = 1.0/3.0*np.sqrt(omega/rovert)
    # else:
    #     C_tau = 1.0
    C_tau = _tausmooth(omega, rovert)

    tau_zt_Rcr = 0.75 * E * C_tau * np.sqrt(1.0 / omega) / rovert

    # reduction factor
    alpha_tau = 0.65  # high fabrifaction quality
    beta_tau = 0.6
    lambda_tau0 = 0.4
    eta_tau = 1.0
    lambda_tau = np.sqrt(sigma_y / np.sqrt(3) / tau_zt_Rcr)

    chi_tau = _buckling_reduction_factor(alpha_tau, beta_tau, eta_tau, lambda_tau0, lambda_tau)

    tau_zt_Rk = chi_tau * sigma_y / np.sqrt(3)
    tau_zt_Rd = tau_zt_Rk / gamma_b

    # buckling interaction parameters

    k_z = 1.25 + 0.75 * chi_z
    k_theta = 1.25 + 0.75 * chi_theta
    k_tau = 1.75 + 0.25 * chi_tau
    k_i = (chi_z * chi_theta) ** 2

    # shell buckling utilization

    utilization = (
        (sigma_z / sigma_z_Rd) ** k_z
        + (sigma_t / sigma_t_Rd) ** k_theta
        - k_i * (sigma_z * sigma_t / sigma_z_Rd / sigma_t_Rd)
        + (tau_zt / tau_zt_Rd) ** k_tau
    )

    return utilization  # this is utilization must be <1


def _buckling_reduction_factor(alpha, beta, eta, lambda_0, lambda_bar):
    """
    Computes a buckling reduction factor used in Eurocode shell buckling formula.
    """

    lambda_p = np.sqrt(alpha / (1.0 - beta))

    ptL = 0.9 * lambda_0
    ptR = 1.1 * lambda_0

    if lambda_bar < ptL:
        chi = 1.0

    elif lambda_bar >= ptL and lambda_bar <= ptR:  # cubic spline section

        fracR = (ptR - lambda_0) / (lambda_p - lambda_0)
        fL = 1.0
        fR = 1 - beta * fracR ** eta
        gL = 0.0
        gR = -beta * eta * fracR ** (eta - 1) / (lambda_p - lambda_0)

        chi = cubic_spline_eval(ptL, ptR, fL, fR, gL, gR, lambda_bar)

    elif lambda_bar > ptR and lambda_bar < lambda_p:
        chi = 1.0 - beta * ((lambda_bar - lambda_0) / (lambda_p - lambda_0)) ** eta

    else:
        chi = alpha / lambda_bar ** 2

    # if (lambda_bar <= lambda_0):
    #     chi = 1.0
    # elif (lambda_bar >= lambda_p):
    #     chi = alpha/lambda_bar**2
    # else:
    #     chi = 1.0 - beta*((lambda_bar-lambda_0)/(lambda_p-lambda_0))**eta

    return chi



import os
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
    import pandas as pd


# The library path is where to find the default input data for LandBOSSE.
ROOT = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../.."))
if ROOT.endswith("wisdem"):
    library_path = os.path.join(ROOT, "library", "landbosse")
else:
    library_path = os.path.join(ROOT, "project_input_template", "project_data")


class OpenMDAODataframeCache:
    """
    This class does not need to be instantiated. This means that the
    cache is shared throughout all parts of the code that needs access
    to any part of the project_data .xlsx files.

    This class is made to read all sheets from xlsx files and store those
    sheets as dictionaries. This is so .xlsx files only need to be parsed
    once.

    One of the use cases for this dataframe cache is in parallel process
    execution using ProcessPoolExecutor. Alternatively, once code use
    the ThreadPoolExecutor (though that wouldn't give the same advantages
    of paralelization).

    Regardless of which executor is used, care must be taken that one thread
    or process cannot mutate the dataframes of another process. So, this
    class make copies of dataframes so the callables running from the
    executor cannot overwrite each other's data.
    """

    # _cache is a class attribute that holds the cache of sheets and their
    # dataframes
    _cache = {}

    @classmethod
    def read_all_sheets_from_xlsx(cls, xlsx_basename, xlsx_path=None):
        """
        If the .xlsx file specified by .xlsx_basename has been read before
        (meaning it is stored as a key on cls._cache), a copy of all the
        dataframes stored under that sheet name is returned. See the note
        about copying in the class docstring for why copies are being made.

        If the xlsx_basename has not been read before, all the sheets are
        read and copies are returned. The sheets are stored on the dictionary
        cache.

        Parameters
        ----------
        xlsx_basename : str
            The base name of the xlsx file to read. This name should
            not include the .xlsx at the end of the filename. This class
            uses XlsxFileOperations to find the dataframes in the
            project_data directory. The xlsx_basename becomes the key
            in the dictionary used to access all the sheets in the
            named .xlsx file.

        xlsx_path : str
            The path from which to read the .xlsx file. This parameter
            has the default value of the library path variable above.

        Returns
        -------
        dict
            A dictionary of dataframes. Keys on the dictionary are names of
            sheets and values in the dictionary are dataframes in that
            .xlsx file.
        """
        if xlsx_basename in cls._cache:
            original = cls._cache[xlsx_basename]
            return cls.copy_dataframes(original)

        if xlsx_path is None:
            xlsx_filename = os.path.join(library_path, f"{xlsx_basename}.xlsx")
        else:
            xlsx_filename = os.path.join(xlsx_path, f"{xlsx_basename}.xlsx")

        xlsx = pd.ExcelFile(xlsx_filename, engine="openpyxl")
        sheets_dict = {sheet_name: xlsx.parse(sheet_name) for sheet_name in xlsx.sheet_names}
        for sheet_name in xlsx.sheet_names:
            sheets_dict[sheet_name].dropna(inplace=True, how="all")
        cls._cache[xlsx_basename] = sheets_dict
        return cls.copy_dataframes(sheets_dict)

    @classmethod
    def copy_dataframes(cls, dict_of_dataframes):
        """
        This copies a dictionary of dataframes. See the class docstring for an
        explanation of why this copying is taking place.

        Parameters
        ----------
        dict_of_dataframes : dict
            The dictionary of dataframes to copy.

        Returns
        -------
        dict
            Keys are the same as the original dictionary of dataframes.
            Values are copies of the origin dataframes.
        """
        return {xlsx_basename: df.copy() for xlsx_basename, df in dict_of_dataframes.items()}



"""Installation strategies for moored floating systems."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import simpy
from marmot import le, process

from wisdem.orbit.core import Vessel, WetStorage
from wisdem.orbit.phases.install import InstallPhase

from .common import TowingGroup, TurbineAssemblyLine, SubstructureAssemblyLine


class MooredSubInstallation(InstallPhase):
    """
    Installation module to model the quayside assembly, tow-out and
    installation at sea of moored substructures.
    """

    phase = "Moored Substructure Installation"
    capex_category = "Substructure"

    #:
    expected_config = {
        "support_vessel": "str",
        "towing_vessel": "str",
        "towing_vessel_groups": {
            "towing_vessels": "int",
            "station_keeping_vessels": "int",
            "num_groups": "int (optional)",
        },
        "substructure": {
            "takt_time": "int | float (optional, default: 0)",
            "towing_speed": "int | float (optional, default: 6 km/h)",
            "unit_cost": "USD",
        },
        "site": {"depth": "m", "distance": "km"},
        "plant": {"num_turbines": "int"},
        "turbine": "dict",
        "port": {
            "sub_assembly_lines": "int (optional, default: 1)",
            "sub_storage": "int (optional, default: inf)",
            "turbine_assembly_cranes": "int (optional, default: 1)",
            "assembly_storage": "int (optional, default: inf)",
            "monthly_rate": "USD/mo (optional)",
            "name": "str (optional)",
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of MooredSubInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : np.array
            Weather data at site.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.setup_simulation(**kwargs)

    def setup_simulation(self, **kwargs):
        """
        Sets up simulation infrastructure.
        - Initializes substructure production
        - Initializes turbine assembly processes
        - Initializes towing groups
        """

        self.distance = self.config["site"]["distance"]
        self.num_turbines = self.config["plant"]["num_turbines"]

        self.initialize_port()
        self.initialize_substructure_production()
        self.initialize_turbine_assembly()
        self.initialize_queue()
        self.initialize_towing_groups()
        self.initialize_support_vessel()

    @property
    def system_capex(self):
        """Returns total procurement cost of the substructures."""

        return self.num_turbines * self.config["substructure"]["unit_cost"]

    def initialize_substructure_production(self):
        """
        Initializes the production of substructures at port. The number of
        independent assembly lines and production time associated with a
        substructure can be configured with the following parameters:

        - self.config["substructure"]["takt_time"]
        - self.config["port"]["sub_assembly_lines"]
        """

        try:
            storage = self.config["port"]["sub_storage"]

        except KeyError:
            storage = float("inf")

        self.wet_storage = WetStorage(self.env, storage)

        try:
            time = self.config["substructure"]["takt_time"]

        except KeyError:
            time = 0

        try:
            lines = self.config["port"]["sub_assembly_lines"]

        except KeyError:
            lines = 1

        to_assemble = [1] * self.num_turbines

        self.sub_assembly_lines = []
        for i in range(lines):
            a = SubstructureAssemblyLine(to_assemble, time, self.wet_storage, i + 1)

            self.env.register(a)
            a.start()
            self.sub_assembly_lines.append(a)

    def initialize_turbine_assembly(self):
        """
        Initializes turbine assembly lines. The number of independent lines
        can be configured with the following parameters:

        - self.config["port"]["turb_assembly_lines"]
        """

        try:
            storage = self.config["port"]["assembly_storage"]

        except KeyError:
            storage = float("inf")

        self.assembly_storage = WetStorage(self.env, storage)

        try:
            lines = self.config["port"]["turbine_assembly_cranes"]

        except KeyError:
            lines = 1

        turbine = self.config["turbine"]
        self.turbine_assembly_lines = []
        for i in range(lines):
            a = TurbineAssemblyLine(self.wet_storage, self.assembly_storage, turbine, i + 1)

            self.env.register(a)
            a.start()
            self.turbine_assembly_lines.append(a)

    def initialize_towing_groups(self, **kwargs):
        """
        Initializes towing groups to bring completed assemblies to site and
        stabilize the assembly during final installation.
        """

        self.installation_groups = []

        vessel = self.config["towing_vessel"]
        num_groups = self.config["towing_vessel_groups"].get("num_groups", 1)
        towing = self.config["towing_vessel_groups"]["towing_vessels"]
        towing_speed = self.config["substructure"].get("towing_speed", 6)

        for i in range(num_groups):
            g = TowingGroup(vessel, num=i + 1)
            self.env.register(g)
            g.initialize()
            self.installation_groups.append(g)

            transfer_moored_substructures_from_storage(
                g,
                self.assembly_storage,
                self.distance,
                self.active_group,
                towing,
                towing_speed,
                **kwargs,
            )

    def initialize_queue(self):
        """
        Initializes the queue, modeled as a ``SimPy.Resource`` that towing
        groups join at site.
        """

        self.active_group = simpy.Resource(self.env, capacity=1)
        self.active_group.vessel = None
        self.active_group.activate = self.env.event()

    def initialize_support_vessel(self, **kwargs):
        """
        Initializes Multi-Purpose Support Vessel to perform installation
        processes at site.
        """

        specs = self.config["support_vessel"]
        vessel = self.initialize_vessel("Multi-Purpose Support Vessel", specs)

        self.env.register(vessel)
        vessel.initialize(mobilize=False)
        self.support_vessel = vessel

        station_keeping_vessels = self.config["towing_vessel_groups"]["station_keeping_vessels"]

        install_moored_substructures(
            self.support_vessel,
            self.active_group,
            self.distance,
            self.num_turbines,
            station_keeping_vessels,
            **kwargs,
        )

    @property
    def detailed_output(self):
        """"""

        return {
            "operational_delays": {
                **{k: self.operational_delay(str(k)) for k in self.sub_assembly_lines},
                **{k: self.operational_delay(str(k)) for k in self.turbine_assembly_lines},
                **{k: self.operational_delay(str(k)) for k in self.installation_groups},
                self.support_vessel: self.operational_delay(str(self.support_vessel)),
            }
        }

    def operational_delay(self, name):
        """"""

        actions = [a for a in self.env.actions if a["agent"] == name]
        delay = sum(a["duration"] for a in actions if "Delay" in a["action"])

        return delay


@process
def transfer_moored_substructures_from_storage(group, feed, distance, queue, towing_vessels, towing_speed, **kwargs):
    """
    Process logic for the towing vessel group.

    Parameters
    ----------
    group : Vessel
        Towing group.
    feed : simpy.Store
        Completed assembly storage.
    distance : int | float
        Distance from port to site.
    towing_vessels : int
        Number of vessels to use for towing to site.
    towing_speed : int | float
        Configured towing speed (km/h)
    """

    towing_time = distance / towing_speed
    transit_time = distance / group.transit_speed

    while True:

        start = group.env.now
        assembly = yield feed.get()
        delay = group.env.now - start

        if delay > 0:
            group.submit_action_log("Delay: No Completed Assemblies Available", delay)

        yield group.group_task(
            "Ballast to Towing Draft",
            6,
            num_vessels=towing_vessels,
            constraints={"windspeed": le(15), "waveheight": le(2.5)},
        )

        yield group.group_task(
            "Tow Substructure",
            towing_time,
            num_vessels=towing_vessels,
            constraints={"windspeed": le(15), "waveheight": le(2.5)},
        )

        # At Site
        with queue.request() as req:
            queue_start = group.env.now
            yield req

            queue_time = group.env.now - queue_start
            if queue_time > 0:
                group.submit_action_log("Queue", queue_time, location="Site")

            queue.vessel = group
            active_start = group.env.now
            queue.activate.succeed()

            # Released by WTIV when objects are depleted
            group.release = group.env.event()
            yield group.release
            active_time = group.env.now - active_start

            queue.vessel = None
            queue.activate = group.env.event()

        yield group.group_task("Transit", transit_time, num_vessels=towing_vessels)


@process
def install_moored_substructures(vessel, queue, distance, substructures, station_keeping_vessels, **kwargs):
    """
    Logic that a Multi-Purpose Support Vessel uses at site to complete the
    installation of moored substructures.

    Parameters
    ----------
    vessel : Vessel
    queue :
    distance : int | float
        Distance between port and site (km).
    substructures : int
        Number of substructures to install before transiting back to port.
    station_keeping_vessels : int
        Number of vessels to use for substructure station keeping during final
        installation at site.
    """

    n = 0
    while n < substructures:
        if queue.vessel:

            start = vessel.env.now
            if n == 0:
                vessel.mobilize()
                yield vessel.transit(distance)

            yield vessel.task_wrapper(
                "Position Substructure",
                2,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )
            yield vessel.task_wrapper(
                "Ballast to Operational Draft",
                6,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )
            yield vessel.task_wrapper(
                "Connect Mooring Lines",
                22,
                suspendable=True,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )
            yield vessel.task_wrapper(
                "Check Mooring Lines",
                12,
                suspendable=True,
                constraints={"windspeed": le(15), "waveheight": le(2.5)},
            )

            group_time = vessel.env.now - start
            queue.vessel.submit_action_log(
                "Positioning Support",
                group_time,
                location="site",
                num_vessels=station_keeping_vessels,
            )
            yield queue.vessel.release.succeed()
            n += 1

        else:
            start = vessel.env.now
            yield queue.activate
            delay_time = vessel.env.now - start

            if n != 0:
                vessel.submit_action_log("Delay", delay_time, location="Site")

    yield vessel.transit(distance)



import os
import sys
import logging
import warnings

import numpy as np
import openmdao.api as om

from wisdem.commonse import fileIO
from wisdem.commonse.mpi_tools import MPI
from wisdem.glue_code.glue_code import WindPark
from wisdem.glue_code.gc_LoadInputs import WindTurbineOntologyPython
from wisdem.glue_code.gc_WT_InitModel import yaml2openmdao
from wisdem.glue_code.gc_PoseOptimization import PoseOptimization

np.warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
# Suppress the maxfev warnings is scipy _minpack_py, line:175
warnings.simplefilter("ignore", RuntimeWarning, lineno=175)

if MPI:
    # from openmdao.api import PetscImpl as impl
    # from mpi4py import MPI
    # from petsc4py import PETSc
    from wisdem.commonse.mpi_tools import subprocessor_loop, subprocessor_stop, map_comm_heirarchical


def run_wisdem(fname_wt_input, fname_modeling_options, fname_opt_options, overridden_values=None, run_only=False):
    # Load all yaml inputs and validate (also fills in defaults)
    wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
    wt_init, modeling_options, opt_options = wt_initial.get_input_data()

    # Initialize openmdao problem. If running with multiple processors in MPI, use parallel finite differencing equal to the number of cores used.
    # Otherwise, initialize the WindPark system normally. Get the rank number for parallelization. We only print output files using the root processor.
    myopt = PoseOptimization(wt_init, modeling_options, opt_options)

    if MPI:

        n_DV = myopt.get_number_design_variables()

        # Extract the number of cores available
        max_cores = MPI.COMM_WORLD.Get_size()

        if max_cores > n_DV:
            raise ValueError(
                "ERROR: please reduce the number of cores, currently set to "
                + str(max_cores)
                + ", to the number of finite differences "
                + str(n_DV)
                + ", which is equal to the number of design variables DV for forward differencing"
                + " and DV times 2 for central differencing,"
                + " or the parallelization logic will not work"
            )

        # Define the color map for the parallelization, determining the maximum number of parallel finite difference (FD) evaluations based on the number of design variables (DV).
        n_FD = min([max_cores, n_DV])

        # Define the color map for the cores
        n_FD = max([n_FD, 1])
        comm_map_down, comm_map_up, color_map = map_comm_heirarchical(n_FD, 1)
        rank = MPI.COMM_WORLD.Get_rank()
        color_i = color_map[rank]
        comm_i = MPI.COMM_WORLD.Split(color_i, 1)
    else:
        color_i = 0
        rank = 0

    folder_output = opt_options["general"]["folder_output"]

    if rank == 0:
        os.makedirs(folder_output, exist_ok=True)

        # create logger
        logger = logging.getLogger("wisdem/weis")
        logger.setLevel(logging.INFO)

        # create handlers
        ht = logging.StreamHandler()
        ht.setLevel(logging.WARNING)

        flog = os.path.join(folder_output, opt_options["general"]["fname_output"] + ".log")
        hf = logging.FileHandler(flog, mode="w")
        hf.setLevel(logging.INFO)

        # create formatters
        formatter_t = logging.Formatter("%(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s")
        formatter_f = logging.Formatter(
            "P%(process)d %(asctime)s %(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s"
        )

        # add formatter to handlers
        ht.setFormatter(formatter_t)
        hf.setFormatter(formatter_f)

        # add handlers to logger
        logger.addHandler(ht)
        logger.addHandler(hf)
        logger.info("Started")

    if color_i == 0:  # the top layer of cores enters
        if MPI:
            # Parallel settings for OpenMDAO
            wt_opt = om.Problem(model=om.Group(num_par_fd=n_FD), comm=comm_i)
            wt_opt.model.add_subsystem(
                "comp", WindPark(modeling_options=modeling_options, opt_options=opt_options), promotes=["*"]
            )
        else:
            # Sequential finite differencing
            wt_opt = om.Problem(model=WindPark(modeling_options=modeling_options, opt_options=opt_options))

        # If at least one of the design variables is active, setup an optimization
        if opt_options["opt_flag"] and not run_only:
            wt_opt = myopt.set_driver(wt_opt)
            wt_opt = myopt.set_objective(wt_opt)
            wt_opt = myopt.set_design_variables(wt_opt, wt_init)
            wt_opt = myopt.set_constraints(wt_opt)
            wt_opt = myopt.set_recorders(wt_opt)

        # Setup openmdao problem
        wt_opt.setup()

        # Load initial wind turbine data from wt_initial to the openmdao problem
        wt_opt = yaml2openmdao(wt_opt, modeling_options, wt_init, opt_options)
        wt_opt = myopt.set_initial(wt_opt, wt_init)

        # If the user provides values in this dict, they overwrite
        # whatever values have been set by the yaml files.
        # This is useful for performing black-box wrapped optimization without
        # needing to modify the yaml files.
        if overridden_values is not None:
            for key in overridden_values:
                wt_opt[key] = overridden_values[key]

        # Place the last design variables from a previous run into the problem.
        # This needs to occur after the above setup() and yaml2openmdao() calls
        # so these values are correctly placed in the problem.
        wt_opt = myopt.set_restart(wt_opt)

        if "check_totals" in opt_options["driver"] and not run_only:
            if opt_options["driver"]["check_totals"]:
                wt_opt.run_model()
                totals = wt_opt.compute_totals()

        if "check_partials" in opt_options["driver"] and not run_only:
            if opt_options["driver"]["check_partials"]:
                wt_opt.run_model()
                checks = wt_opt.check_partials(compact_print=True)

        sys.stdout.flush()

        if opt_options["driver"]["step_size_study"]["flag"] and not run_only:
            wt_opt.run_model()
            study_options = opt_options["driver"]["step_size_study"]
            step_sizes = study_options["step_sizes"]
            all_derivs = {}
            for idx, step_size in enumerate(step_sizes):
                wt_opt.model.approx_totals(method="fd", step=step_size, form=study_options["form"])

                if study_options["of"]:
                    of = study_options["of"]
                else:
                    of = None

                if study_options["wrt"]:
                    wrt = study_options["wrt"]
                else:
                    wrt = None

                derivs = wt_opt.compute_totals(of=of, wrt=wrt, driver_scaling=study_options["driver_scaling"])
                all_derivs[idx] = derivs
                all_derivs[idx]["step_size"] = step_size
            np.save("total_derivs.npy", all_derivs)

        # Run openmdao problem
        elif opt_options["opt_flag"] and not run_only:
            wt_opt.run_driver()
        else:
            wt_opt.run_model()

        if (not MPI) or (MPI and rank == 0):
            # Save data coming from openmdao to an output yaml file
            froot_out = os.path.join(folder_output, opt_options["general"]["fname_output"])
            wt_initial.write_ontology(wt_opt, froot_out)
            wt_initial.write_options(froot_out)

            # Save data to numpy and matlab arrays
            fileIO.save_data(froot_out, wt_opt)

    if rank == 0:
        return wt_opt, modeling_options, opt_options
    else:
        return [], [], []


def load_wisdem(frootin):
    froot = os.path.splitext(frootin)[0]
    fgeom = froot + ".yaml"
    fmodel = froot + "-modeling.yaml"
    fopt = froot + "-analysis.yaml"
    fpkl = froot + ".pkl"

    # Load all yaml inputs and validate (also fills in defaults)
    wt_initial = WindTurbineOntologyPython(fgeom, fmodel, fopt)
    wt_init, modeling_options, opt_options = wt_initial.get_input_data()

    wt_opt = om.Problem(model=WindPark(modeling_options=modeling_options, opt_options=opt_options))
    wt_opt.setup()

    wt_opt = fileIO.load_data(fpkl, wt_opt)

    return wt_opt, modeling_options, opt_options



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

from numpy.testing import assert_almost_equal

from wisdem.orbit import ProjectManager
from wisdem.orbit.core.library import extract_library_specs

fixed = extract_library_specs("config", "complete_project")
floating = extract_library_specs("config", "complete_floating_project")


def test_fixed_phase_cost_passing():

    project = ProjectManager(fixed)
    project.run()

    assert_almost_equal(
        project.phases["MonopileDesign"].total_cost,
        project.phases["MonopileInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["ScourProtectionDesign"].total_cost,
        project.phases["ScourProtectionInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["ArraySystemDesign"].total_cost,
        project.phases["ArrayCableInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["ExportSystemDesign"].total_cost,
        project.phases["ExportCableInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["OffshoreSubstationDesign"].total_cost,
        project.phases["OffshoreSubstationInstallation"].system_capex,
    )


def test_floating_phase_cost_passing():

    project = ProjectManager(floating)
    project.run()

    assert_almost_equal(
        project.phases["MooringSystemDesign"].total_cost,
        project.phases["MooringSystemInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["SemiSubmersibleDesign"].total_cost,
        project.phases["MooredSubInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["ArraySystemDesign"].total_cost,
        project.phases["ArrayCableInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["ExportSystemDesign"].total_cost,
        project.phases["ExportCableInstallation"].system_capex,
    )

    assert_almost_equal(
        project.phases["OffshoreSubstationDesign"].total_cost,
        project.phases["OffshoreSubstationInstallation"].system_capex,
    )

    spar = deepcopy(floating)
    spar["design_phases"].remove("SemiSubmersibleDesign")
    spar["design_phases"].append("SparDesign")

    project = ProjectManager(spar)
    project.run()

    assert_almost_equal(
        project.phases["SparDesign"].total_cost,
        project.phases["MooredSubInstallation"].system_capex,
    )



import unittest

import numpy as np
import numpy.testing as npt
from wisdem.pyframe3dd import Frame, Options, NodeData, ElementData, ReactionData, StaticLoadCase


class TestBreakdown(unittest.TestCase):
    def testRXNs(self):
        # Nodal data
        ntop = 5
        nbase = 10
        nall = ntop + nbase - 1
        xyz_top = np.c_[np.linspace(0.0, -10, ntop), np.zeros(ntop), 100 * np.ones(ntop)]
        xyz_base = np.c_[np.zeros((nbase, 2)), np.linspace(0, 100, nbase)]
        xyz_all = np.vstack((xyz_base, xyz_top[1:, :]))

        zero_top = np.zeros(ntop)
        inode_top = np.arange(ntop, dtype=np.int_) + 1
        node_top = NodeData(inode_top, xyz_top[:, 0], xyz_top[:, 1], xyz_top[:, 2], zero_top)

        zero_base = np.zeros(nbase)
        inode_base = np.arange(nbase, dtype=np.int_) + 1
        node_base = NodeData(inode_base, xyz_base[:, 0], xyz_base[:, 1], xyz_base[:, 2], zero_base)

        zero_all = np.zeros(nall)
        inode_all = np.arange(nall, dtype=np.int_) + 1
        node_all = NodeData(inode_all, xyz_all[:, 0], xyz_all[:, 1], xyz_all[:, 2], zero_all)

        # Reactions
        rnode = np.array([1], dtype=np.int_)
        Kval = np.array([1], dtype=np.int_)
        rigid = 1
        reactions = ReactionData(rnode, Kval, Kval, Kval, Kval, Kval, Kval, rigid)

        # Element data
        top1 = np.ones(ntop - 1)
        base1 = np.ones(nbase - 1)
        all1 = np.ones(nall - 1)
        ielem_top = np.arange(1, ntop)
        ielem_base = np.arange(1, nbase)
        ielem_all = np.arange(1, nall)
        N1_top = np.arange(ntop - 1, dtype=np.int_) + 1
        N1_base = np.arange(nbase - 1, dtype=np.int_) + 1
        N1_all = np.arange(nall - 1, dtype=np.int_) + 1
        N2_top = N1_top + 1
        N2_base = N1_base + 1
        N2_all = N1_all + 1
        Ax = 2.0
        Asx = Asy = 4.0
        J0 = 2e1
        Ix = Iy = 1e1
        E = 5e6
        G = 5e6
        density = 8e2
        elem_top = ElementData(
            ielem_top,
            N1_top,
            N2_top,
            Ax * top1,
            Asx * top1,
            Asy * top1,
            J0 * top1,
            Ix * top1,
            Iy * top1,
            E * top1,
            G * top1,
            0.0 * top1,
            density * top1,
        )
        elem_base = ElementData(
            ielem_base,
            N1_base,
            N2_base,
            Ax * base1,
            Asx * base1,
            Asy * base1,
            J0 * base1,
            Ix * base1,
            Iy * base1,
            E * base1,
            G * base1,
            0.0 * base1,
            density * base1,
        )
        elem_all = ElementData(
            ielem_all,
            N1_all,
            N2_all,
            Ax * all1,
            Asx * all1,
            Asy * all1,
            J0 * all1,
            Ix * all1,
            Iy * all1,
            E * all1,
            G * all1,
            0.0 * all1,
            density * all1,
        )

        # parameters
        shear = False  # 1: include shear deformation
        geom = False  # 1: include geometric stiffness
        dx = -1.0  # x-axis increment for internal forces
        options = Options(shear, geom, dx)

        frame_top = Frame(node_top, reactions, elem_top, options)
        frame_base = Frame(node_base, reactions, elem_base, options)
        frame_all = Frame(node_all, reactions, elem_all, options)

        # load case
        gx = 0.0
        gy = 0.0
        gz = -9.81
        load_top = StaticLoadCase(gx, gy, gz)
        load_base = StaticLoadCase(gx, gy, gz)
        load_all = StaticLoadCase(gx, gy, gz)

        # pseudo-rotor loads
        nF_top = [inode_top[-1]]
        nF_base = [inode_base[-1]]
        nF_all = [inode_all[-1]]
        F = 1e2
        M = 1e3
        load_top.changePointLoads(nF_top, [F], [F], [F], [M], [M], [M])
        load_all.changePointLoads(nF_all, [F], [F], [F], [M], [M], [M])

        frame_top.addLoadCase(load_top)
        frame_all.addLoadCase(load_all)

        # Added mass
        AN_top = [inode_top[-2]]
        AN_all = [inode_all[-2]]
        AN_base = [inode_base[-2]]
        EMs = np.array([1e2])
        EMxx = EMyy = EMzz = EMxy = EMxz = EMyz = np.array([0.0])
        rhox = rhoy = rhoz = np.array([0.0])
        addGravityLoad = True
        frame_top.changeExtraNodeMass(AN_top, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)
        frame_all.changeExtraNodeMass(AN_all, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)

        # Run first models
        disp_top, forces_top, rxns_top, _, _, _ = frame_top.run()
        disp_all, forces_all, rxns_all, _, _, _ = frame_all.run()

        # Transfer loads to base
        load_base.changePointLoads(
            nF_base,
            [-rxns_top.Fx],
            [-rxns_top.Fy],
            [-rxns_top.Fz],
            [-rxns_top.Mxx],
            [-rxns_top.Myy],
            [-rxns_top.Mzz],
        )
        frame_base.addLoadCase(load_base)
        disp_base, forces_base, rxns_base, _, _, _ = frame_base.run()

        npt.assert_almost_equal(rxns_all.Fx, rxns_base.Fx, decimal=3)
        npt.assert_almost_equal(rxns_all.Fy, rxns_base.Fy, decimal=3)
        npt.assert_almost_equal(rxns_all.Fz, rxns_base.Fz, decimal=3)
        npt.assert_almost_equal(rxns_all.Mxx, rxns_base.Mxx, decimal=2)
        npt.assert_almost_equal(rxns_all.Myy, rxns_base.Myy, decimal=2)
        npt.assert_almost_equal(rxns_all.Mzz, rxns_base.Mzz, decimal=2)

    def testRXNs2(self):
        # Nodal data
        ntop = 5
        nbase = 10
        nall = ntop + nbase - 1
        xyz_top = np.c_[np.linspace(0.0, -10, ntop), np.zeros(ntop), 100 * np.ones(ntop)]
        xyz_base = np.c_[np.zeros((nbase, 2)), np.linspace(0, 100, nbase)]
        xyz_all = np.vstack((xyz_base, xyz_top[1:, :]))

        zero_top = np.zeros(ntop)
        inode_top = np.arange(ntop, dtype=np.int_) + 1
        node_top = NodeData(inode_top, xyz_top[:, 0], xyz_top[:, 1], xyz_top[:, 2], zero_top)

        zero_base = np.zeros(nbase)
        inode_base = np.arange(nbase, dtype=np.int_) + 1
        node_base = NodeData(inode_base, xyz_base[:, 0], xyz_base[:, 1], xyz_base[:, 2], zero_base)

        zero_all = np.zeros(nall)
        inode_all = np.arange(nall, dtype=np.int_) + 1
        node_all = NodeData(inode_all, xyz_all[:, 0], xyz_all[:, 1], xyz_all[:, 2], zero_all)

        # Reactions
        rnode = np.array([1], dtype=np.int_)
        Kval = np.array([1], dtype=np.int_)
        rigid = 1
        reactions = ReactionData(rnode, Kval, Kval, Kval, Kval, Kval, Kval, rigid)

        # Element data
        top1 = np.ones(ntop - 1)
        base1 = np.ones(nbase - 1)
        all1 = np.ones(nall - 1)
        ielem_top = np.arange(1, ntop)
        ielem_base = np.arange(1, nbase)
        ielem_all = np.arange(1, nall)
        N1_top = np.arange(ntop - 1, dtype=np.int_) + 1
        N1_base = np.arange(nbase - 1, dtype=np.int_) + 1
        N1_all = np.arange(nall - 1, dtype=np.int_) + 1
        N2_top = N1_top + 1
        N2_base = N1_base + 1
        N2_all = N1_all + 1
        Ax = 2.0
        Asx = Asy = 4.0
        J0 = 2e1
        Ix = Iy = 1e1
        E = 5e6
        G = 5e6
        density = 8e2
        elem_top = ElementData(
            ielem_top,
            N1_top,
            N2_top,
            Ax * top1,
            Asx * top1,
            Asy * top1,
            J0 * top1,
            Ix * top1,
            Iy * top1,
            E * top1,
            G * top1,
            0.0 * top1,
            density * top1,
        )
        elem_base = ElementData(
            ielem_base,
            N1_base,
            N2_base,
            Ax * base1,
            Asx * base1,
            Asy * base1,
            J0 * base1,
            Ix * base1,
            Iy * base1,
            E * base1,
            G * base1,
            0.0 * base1,
            density * base1,
        )
        elem_all = ElementData(
            ielem_all,
            N1_all,
            N2_all,
            Ax * all1,
            Asx * all1,
            Asy * all1,
            J0 * all1,
            Ix * all1,
            Iy * all1,
            E * all1,
            G * all1,
            0.0 * all1,
            density * all1,
        )

        # parameters
        shear = False  # 1: include shear deformation
        geom = False  # 1: include geometric stiffness
        dx = -1.0  # x-axis increment for internal forces
        options = Options(shear, geom, dx)

        frame_top = Frame(node_top, reactions, elem_top, options)
        frame_base = Frame(node_base, reactions, elem_base, options)
        frame_all = Frame(node_all, reactions, elem_all, options)

        # dynamics
        nM = 20  # number of desired dynamic modes of vibration
        Mmethod = 1  # 1: subspace Jacobi     2: Stodola
        lump = 0  # 0: consistent mass ... 1: lumped mass matrix
        tol = 1e-7  # mode shape tolerance
        shift = 0.0  # shift value ... for unrestrained structures
        frame_top.enableDynamics(nM, Mmethod, lump, tol, shift)
        frame_base.enableDynamics(nM, Mmethod, lump, tol, shift)
        frame_all.enableDynamics(nM, Mmethod, lump, tol, shift)

        # load case
        gx = 0.0
        gy = 0.0
        gz = -9.81
        load_top = StaticLoadCase(gx, gy, gz)
        load_base = StaticLoadCase(gx, gy, gz)
        load_all = StaticLoadCase(gx, gy, gz)

        # pseudo-rotor loads
        nF_top = [inode_top[-1]]
        nF_base = [inode_base[-1]]
        nF_all = [inode_all[-1]]
        F = 1e2
        M = 1e3
        load_top.changePointLoads(nF_top, [F], [F], [F], [M], [M], [M])
        load_all.changePointLoads(nF_all, [F], [F], [F], [M], [M], [M])

        frame_top.addLoadCase(load_top)
        frame_all.addLoadCase(load_all)

        # Added mass
        AN_top = [inode_top[-2]]
        AN_all = [inode_all[-2]]
        AN_base = [inode_base[-2]]
        EMs = np.array([1e2])
        EMxx = EMyy = EMzz = EMxy = EMxz = EMyz = np.array([0.0])
        rhox = rhoy = rhoz = np.array([0.0])
        addGravityLoad = True
        frame_top.changeExtraNodeMass(AN_top, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)
        frame_all.changeExtraNodeMass(AN_all, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)

        # Run first models
        disp_top, forces_top, rxns_top, _, _, modal_top = frame_top.run()
        disp_all, forces_all, rxns_all, _, _, modal_all = frame_all.run()

        # Transfer loads to base
        load_base.changePointLoads(
            nF_base,
            [-rxns_top.Fx],
            [-rxns_top.Fy],
            [-rxns_top.Fz],
            [-rxns_top.Mxx],
            [-rxns_top.Myy],
            [-rxns_top.Mzz],
        )
        frame_base.addLoadCase(load_base)
        disp_base, forces_base, rxns_base, _, _, modal_base = frame_base.run()

        # npt.assert_almost_equal(rxns_all.Fx, rxns_base.Fx, decimal=3)
        # npt.assert_almost_equal(rxns_all.Fy, rxns_base.Fy, decimal=3)
        # npt.assert_almost_equal(rxns_all.Fz, rxns_base.Fz, decimal=3)
        # npt.assert_almost_equal(rxns_all.Mxx, rxns_base.Mxx, decimal=2)
        # npt.assert_almost_equal(rxns_all.Myy, rxns_base.Myy, decimal=2)
        # npt.assert_almost_equal(rxns_all.Mzz, rxns_base.Mzz, decimal=2)
        print(modal_top.freq)
        print(modal_base.freq)
        print(modal_all.freq)

        """
        newk = np.array([rxns_top.Fx.flatten() / disp_base.dx.flatten()[-1],
                         rxns_top.Fy.flatten() / disp_base.dy.flatten()[-1],
                         rxns_top.Fz.flatten() / disp_base.dz.flatten()[-1],
                         rxns_top.Mxx.flatten() / disp_base.dxrot.flatten()[-1],
                         rxns_top.Myy.flatten() / disp_base.dzrot.flatten()[-1],
                         rxns_top.Mzz.flatten() / disp_base.dyrot.flatten()[-1] ])
        newk = np.abs(newk).flatten()
        print(newk)
        newk = 1e4*np.ones(6)
        new_react = ReactionData(rnode, newk[0], newk[1], newk[2], newk[3], newk[4], newk[5], rigid)
        frame_top = Frame(node_top, new_react, elem_top, options)
        frame_top.enableDynamics(nM, Mmethod, lump, tol, shift)
        frame_top.addLoadCase(load_top)
        frame_top.changeExtraNodeMass(AN_top, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)
        disp_top2, forces_top2, rxns_top2, _, _, modal_top2 = frame_top.run()
        print(disp_top.dx.flatten()[-1], disp_all.dx.flatten()[-1], disp_top2.dx.flatten()[-1])
        """

    def testModal(self):

        # nodes
        nn = 11
        node = np.arange(1, nn + 1, dtype=np.int_)
        x = y = r = np.zeros(nn)
        z = np.linspace(0, 100, nn)
        nodes = NodeData(node, x, y, z, r)

        # reactions
        rnode = np.array([1], dtype=np.int_)
        Kx = Ky = Kz = Ktx = Kty = Ktz = np.ones(1)
        rigid = 1
        reactions = ReactionData(rnode, Kx, Ky, Kz, Ktx, Kty, Ktz, rigid)

        # elements
        ne = nn - 1
        EL = np.arange(1, ne + 1, dtype=np.int_)
        N1 = np.arange(1, nn, dtype=np.int_)
        N2 = np.arange(2, nn + 1, dtype=np.int_)
        Ax = Jx = Iy = Iz = 10 * np.ones(ne)
        Asy = Asz = 8 * np.ones(ne)
        E = 2e6 * np.ones(ne)
        G = 1e6 * np.ones(ne)
        roll = np.zeros(ne)
        rho = 1e-5 * np.ones(ne)
        elements = ElementData(EL, N1, N2, Ax, Asy, Asz, Jx, Iy, Iz, E, G, roll, rho)

        # parameters
        shear = False  # 1: include shear deformation
        geom = True  # 1: include geometric stiffness
        dx = -1.0  # x-axis increment for internal forces
        options = Options(shear, geom, dx)

        #### Fixed-free
        frame_mass = Frame(nodes, reactions, elements, options)
        frame_force = Frame(nodes, reactions, elements, options)

        # dynamics
        nM = 15  # number of desired dynamic modes of vibration
        Mmethod = 1  # 1: subspace Jacobi     2: Stodola
        lump = 0  # 0: consistent mass ... 1: lumped mass matrix
        tol = 1e-7  # mode shape tolerance
        shift = -1e2  # shift value ... for unrestrained structures
        frame_mass.enableDynamics(nM, Mmethod, lump, tol, shift)
        frame_force.enableDynamics(nM, Mmethod, lump, tol, shift)

        # load cases 1
        gx = 0.0
        gy = 0.0
        gz = -980.6

        load_mass = StaticLoadCase(gx, gy, gz)
        load_force = StaticLoadCase(gx, gy, gz)

        mymass = 4.0

        # pseudo-mass loads
        nadd = [node[-2]]
        F = mymass * gz
        load_force.changePointLoads(nadd, [0], [0], [F], [0], [0], [0])
        frame_force.addLoadCase(load_force)

        # Added mass
        frame_mass.addLoadCase(load_mass)
        EMs = [mymass]
        EMxx = EMyy = EMzz = EMxy = EMxz = EMyz = np.array([0.0])
        rhox = rhoy = rhoz = np.array([0.0])
        addGravityLoad = True
        frame_mass.changeExtraNodeMass(nadd, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)
        frame_force.changeExtraNodeMass(nadd, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, False)

        _, _, rxns_mass, _, _, modal_mass = frame_mass.run()
        _, _, rxns_force, _, _, modal_force = frame_force.run()

        npt.assert_almost_equal(modal_mass.freq, modal_force.freq)

        npt.assert_almost_equal(rxns_mass.Fx, rxns_force.Fx)
        npt.assert_almost_equal(rxns_mass.Fy, rxns_force.Fy)
        npt.assert_almost_equal(rxns_mass.Fz, rxns_force.Fz)
        npt.assert_almost_equal(rxns_mass.Mxx, rxns_force.Mxx)
        npt.assert_almost_equal(rxns_mass.Myy, rxns_force.Myy)
        npt.assert_almost_equal(rxns_mass.Mzz, rxns_force.Mzz)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestBreakdown))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import numpy as np


class LineType:
    """A class to hold the various properties of a mooring line type"""

    def __init__(self, name, d, massden, EA, MBL=0.0, cost=0.0, notes="", input_d=0.0, input_type=""):
        """Initialize LineType attributes

        Parameters
        ----------
        name : string
            identifier string
        d : float
            volume-equivalent diameter [m]
        massden : float
            linear mass density [kg/m] used to calculate weight density (w) [N/m]
        EA : float
            extensional stiffness [N]
        MBL : float, optional
            Minimum breaking load [N]. The default is 0.0.
        cost : float, optional
            material cost per unit length [$/m]. The default is 0.0.
        notes : string, optional
            optional notes/description string. The default is "".
        input_d : float, optional
            the given input diameter that has not been adjusted for the line's volume [m]. The default is 0.0.
        input_type : string, optional
            the type of the line (e.g. chain, polyester), different from the name. The default is "".
        Returns
        -------
        None.

        """
        self.name = name  # identifier string
        self.d = d  # volume-equivalent diameter [m]
        self.m = massden  # linear desnity [kg/m]
        self.w = (massden - np.pi / 4 * d * d * 1025) * 9.81  # wet weight [N/m]
        self.EA = EA  # extensional stiffness [N]
        self.MBL = MBL  # minimum breaking load [N]
        self.cost = cost  # material cost of line per unit length [$/m]
        self.notes = notes  # optional notes/description string
        self.input_d = input_d  # the non-volume-equivalent, input diameter [m]
        self.input_type = input_type  # line type string (e.g. chain, polyester)
